#!/usr/bin/env python
# -*- coding: utf-8 -*-

from ..output import stat_functions as stat
from ..output import stat_dist
from . import constraints
from .. import likelihood as logl
from . import direction

import numpy as np


EPS=3.0e-16 
TOLX=(4*EPS) 
STPMX=100.0 



class Computation:
  def __init__(self,args, panel, gtol, tolx, callback, nummerical, diag_hess):
    """callback is a function taking any named arguments """
    if callback is None:
      callback = lambda **kwargs: None#function that does nothing
    self.callback = callback
    self.gradient=logl.gradient(panel, self.callback)
    self.gtol = panel.options.tolerance.value
    self.tolx = tolx
    self.hessian=logl.hessian(panel,self.gradient, self.callback)
    self.panel=panel
    self.constr=None
    self.CI=0
    self.weak_mc_dict={}
    self.mc_problems=[]
    self.H_correl_problem=False
    self.singularity_problems=False
    self.num_hess_count = 0
    self.H, self.g, self.G = None, None, None
    self.mcollcheck = False
    self.rec =[]
    self.nummerical = nummerical
    self.quit = False
    self.avg_incr = 0

    self.diag_hess = diag_hess
    self.CI_anal = 2
    p, q, d, k, m = panel.pqdkm
    self.init_arma_its = 0
    self.set_constr(args)




  def set(self, its,increment,lmbda,rev, add_dyn_constr, H, ll, x):
    self.its=its
    self.lmbda=lmbda
    self.has_reversed_directions=rev
    self.increment=increment
    self.constr = None

    self.constr_old=self.constr
    self.constr = constraints.Constraints(self.panel,x,its)

    if self.constr is None:
      self.H_correl_problem,self.mc_problems,self.weak_mc_dict=False, [],{}
      self.singularity_problems=(len(self.mc_problems)>0) or self.H_correl_problem
      return

    if self.panel.options.constraints_engine.value:
      self.constr.add_static_constraints(self.panel, its, ll,  np.nonzero(ll.var<ll.minvar)[1])	

      self.constr.add_dynamic_constraints(self, H, ll)	

    self.CI=self.constr.CI
    self.H_correl_problem=self.constr.H_correl_problem	
    self.mc_problems=self.constr.mc_problems
    self.weak_mc_dict=self.constr.weak_mc_dict


  def exec(self, dx_realized,hessin, H, incr, its, ls, calc = True):
    f, x, g_old, rev, alam,ll = ls.f, ls.x, ls.g, ls.rev, ls.alam, ls.ll
    #Thhese setting may not hold for all circumstances, and should be tested properly:

    NUM_ITER = 5
    TOTP_TOL = 1e-15


    g, G = self.calc_gradient(ll)
    dx, dx_norm, H_ = direction.get(g, x, H, self.constr, f, hessin, simple=False)
    
    a = np.ones(len(g))
    if not self.constr is None:
      a[list(self.constr.fixed.keys())] =0		
      a[ls.applied_constraints] = 0    

    pgain, totpgain = potential_gain(dx*a, g, H)
    self.totpgain = totpgain
    max_pgain = abs(max(pgain))


    g_norm =np.max(np.abs(g*a*x)/(abs(f)+1e-12) )
    gtol = self.gtol
    if sum(a)==1:
      gtol = 1e-10
    self.rec.append(str(i) for i in [totpgain, max_pgain, incr, g_norm])
    CI=0
    if not self.CI is None:
      CI = self.CI

    det = np.linalg.det(H)
    se = [None]*len(H)	

    if (ls.conv == 3) or (its >NUM_ITER and ((abs(g_norm) < gtol) or (abs(totpgain)<TOTP_TOL))
        or its>=self.panel.options.max_iterations.value):
      return self.handle_convergence(ll, g, H, x, f, hessin, totpgain, its, TOTP_TOL, ls, g_norm, se, G)
    if not self.panel.options.supress_output.value:
      print(f"its:{its}, f:{f}, gnorm: {abs(g_norm)}")


    self.avg_incr = incr + self.avg_incr*0.7
    self.ev_constr = False#self.CI>1000

    err = np.max(np.abs(dx_realized)) < 100*TOLX

    analytic_calc = (err and self.num_hess_count>4) or ((self.CI>10000) and (self.num_hess_count>5) and abs(g_norm)>10*self.gtol)#or (self.num_hess_count> and its <30) #or ((self.CI>10000) and (self.num_hess_count>1))
    analytic_calc = analytic_calc and (self.panel.options.use_analytical.value>0)
    if calc:
      if analytic_calc:
        H = self.calc_hessian(ll)
        a = 0.5
        try:
          hessin = np.linalg.inv(H)*a + (1-a)*hessin
        except:
          pass
        self.num_hess_count = 0

      else:
        self.num_hess_count +=1
        try:
          hessin=self.hessin_num(hessin, g-g_old, dx_realized)
          Hn = np.linalg.inv(hessin)
          H = Hn
        except:
          H = self.calc_hessian(ll)

    self.H, self.g, self.G = H, g, G

    self.set(its, incr, alam, rev, True, H, ll, x)



    return x, f, hessin, H, g, 0, se, det, analytic_calc
  
  def set_constr(self, args):
    self.constr = constraints.Constraints(self.panel, args)
    self.constr.add_static_constraints(self.panel, 0)	    

  def handle_convergence(self, ll, g, H, x, f, hessin, totpgain, its, TOTP_TOL, ls, g_norm, se, G):
    Ha = self.calc_hessian(ll)    
    keep = [True]*len(H)
    if not self.constr is None:      
      self.constr.add_dynamic_constraints(self, Ha, ll,ll.args.args_v)
      keep = [not i in self.constr.fixed.keys() for i in range(len(H))]
    Ha_keep = Ha[keep][:,keep]
    self.CI_anal = condition_index(Ha_keep)
    try:
      det = np.linalg.det(Ha_keep)
      hessin[keep][:,keep] = np.linalg.inv(Ha_keep)
      se = stat.robust_se(self.panel, 100, hessin, G)[0]				
    except Exception as e:
      print(e)

    if abs(g_norm) < self.gtol:
      return x, f, hessin, Ha, g, 1, se, det, True
    elif abs(totpgain)<TOTP_TOL:
      return x, f, hessin, Ha, g, 2, se, det, True
    elif its>=self.panel.options.max_iterations.value:
      return x, f, hessin, Ha, g, 3, se, det, True
    elif (ls.conv == 2) :
      return x, f, hessin, Ha, g, 4, se, det, True  
    elif (ls.conv == 3) :
      return x, f, hessin, Ha, g, 4, se, det, True      
    
  def calc_gradient(self,ll):
    dLL_lnv, DLL_e=logl.func_gradent(ll,self.panel)
    self.LL_gradient_tobit(ll, DLL_e, dLL_lnv)
    g, G = self.gradient.get(ll,DLL_e,dLL_lnv,return_G=True)
    return g, G


  def calc_hessian(self, ll):
    d2LL_de2, d2LL_dln_de, d2LL_dln2 = logl.func_hessian(ll,self.panel)
    self.LL_hessian_tobit(ll, d2LL_de2, d2LL_dln_de, d2LL_dln2)
    H = self.hessian.get(ll,d2LL_de2,d2LL_dln_de,d2LL_dln2)	
    return H

  def LL_gradient_tobit(self,ll,DLL_e,dLL_lnv):
    g=[1,-1]
    self.f=[None,None]
    self.f_F=[None,None]
    for i in [0,1]:
      if self.panel.tobit_active[i]:
        I=self.panel.tobit_I[i]
        self.f[i]=stat_dist.norm(g[i]*ll.e_norm[I], cdf = False)
        self.f_F[i]=(ll.F[i]!=0)*self.f[i]/(ll.F[i]+(ll.F[i]==0))
        self.v_inv05=ll.v_inv**0.5
        DLL_e[I]=g[i]*self.f_F[i]*self.v_inv05[I]
        dLL_lnv[I]=-0.5*DLL_e[I]*ll.e_RE[I]
        a=0


  def LL_hessian_tobit(self,ll,d2LL_de2,d2LL_dln_de,d2LL_dln2):
    g=[1,-1]
    if sum(self.panel.tobit_active)==0:
      return
    self.f=[None,None]
    e1s1=ll.e_norm
    e2s2=ll.e_REsq*ll.v_inv
    e3s3=e2s2*e1s1
    e1s2=e1s1*self.v_inv05
    e1s3=e1s1*ll.v_inv
    e2s3=e2s2*self.v_inv05
    f_F=self.f_F
    for i in [0,1]:
      if self.panel.tobit_active[i]:
        I=self.panel.tobit_I[i]
        f_F2=self.f_F[i]**2
        d2LL_de2[I]=      -(g[i]*f_F[i]*e1s3[I] + f_F2*ll.v_inv[I])
        d2LL_dln_de[I] =   0.5*(f_F2*e1s2[I]  +  g[i]*f_F[i]*(e2s3[I]-self.v_inv05[I]))
        d2LL_dln2[I] =     0.25*(f_F2*e2s2[I]  +  g[i]*f_F[i]*(e1s1[I]-e3s3[I]))


  def hessin_num(self, hessin, dg, xi):				#Compute difference of gradients,
    n=len(dg)
    #and difference times current matrix:
    hdg=(np.dot(hessin,dg.reshape(n,1))).flatten()
    fac=fae=sumdg=sumxi=0.0 							#Calculate dot products for the denominators. 
    fac = np.sum(dg*xi) 
    fae = np.sum(dg*hdg)
    sumdg = np.sum(dg*dg) 
    sumxi = np.sum(xi*xi) 
    if (fac < (EPS*sumdg*sumxi)**0.5):  					#Skip update if fac not sufficiently positive.
      fac=1.0/fac 
      fad=1.0/fae 
                              #The vector that makes BFGS different from DFP:
      dg=fac*xi-fad*hdg   
      #The BFGS updating formula:
      hessin+=fac*xi.reshape(n,1)*xi.reshape(1,n)
      hessin-=fad*hdg.reshape(n,1)*hdg.reshape(1,n)
      hessin+=fae*dg.reshape(n,1)*dg.reshape(1,n)		

    return hessin



def det_managed(H):
  try:
    return np.linalg.det(H)
  except:
    return 1e+100

def inv_hess(hessian):
  try:
    h=-np.linalg.inv(hessian)
  except:
    return None	
  return h

def condition_index(H):
  n = len(H)
  d=np.maximum(np.abs(np.diag(H)).reshape((n,1)),1e-30)**0.5
  C = -H/(d*d.T)
  ev = np.abs(np.linalg.eigvals(C))**0.5
  if min(ev) == 0:
    return 1e-150
  return max(ev)/min(ev)	


def hess_inv(h, hessin):
  try:
    h_inv = np.linalg.inv(h)
  except Exception as e:
    print(e)
    return hessin
  return h_inv



def potential_gain(dx, g, H):
  """Returns the potential gain of including each variables, given that all other variables are included and that the 
  quadratic model is correct. An alternative convercence criteria"""
  n=len(dx)
  rng=np.arange(len(dx))
  dxLL=dx*0
  dxLL_full=(sum(g*dx)+0.5*np.dot(dx.reshape((1,n)),
                                  np.dot(H,dx.reshape((n,1)))
                                  ))[0,0]
  for i in range(len(dx)):
    dxi=dx*(rng!=i)
    dxLL[i]=dxLL_full-(sum(g*dxi)+0.5*np.dot(dxi.reshape((1,n)),np.dot(H,dxi.reshape((n,1)))))[0,0]

  return dxLL, dxLL_full






#!/usr/bin/env python
# -*- coding: utf-8 -*-

#contains the log likelihood object

from ..output import stat_functions
from .. import random_effects as re
from .. import functions as fu
from . import function
from ..output import stat_dist
from ..processing import model_parser
from . import arma

import numpy as np
import traceback
import time


class LL:
  """Calculates the log likelihood given arguments arg (either in dictonary or array form), and creates an object
  that store dynamic variables that depend on the \n
  If args is a dictionary, the ARMA-GARCH orders are 
  determined from the dictionary. If args is a vector, the ARMA-GARCH order needs to be consistent
  with the  panel object
  """
  def __init__(self,args,panel,constraints=None,print_err=False):
    self.err_msg=''
    self.errmsg_h=''

    #checking settings. If the FE/RE is done on the data before LL
    gfre=panel.options.fixed_random_group_eff.value
    tfre=panel.options.fixed_random_time_eff.value
    vfre=panel.options.fixed_random_variance_eff.value

    self.re_obj_i=re.re_obj(panel,True,panel.T_i,panel.T_i,gfre)
    self.re_obj_t=re.re_obj(panel,False,panel.date_count_mtrx,panel.date_count,tfre)
    self.re_obj_i_v=re.re_obj(panel,True,panel.T_i,panel.T_i,gfre*vfre)
    self.re_obj_t_v=re.re_obj(panel,False,panel.date_count_mtrx,panel.date_count,tfre*vfre)

    self.args=panel.args.create_args(args,panel,constraints)
    self.h_err=""
    self.LL=None
    #self.LL=self.LL_calc(panel)
    try:
      self.LL=self.LL_calc(panel)
      if np.isnan(self.LL):
        self.LL=None						
    except Exception as e:
      if print_err:
        traceback.print_exc()
        print(self.errmsg_h)




  def LL_calc(self,panel):
    X=panel.XIV
    N, T, k = X.shape
    incl = panel.included[3]
    self.set_var_bounds(panel)
    
    G = fu.dot(panel.W_a, self.args.args_d['omega'])
    if 'initvar' in self.args.args_d:
      G[0,0,0] = abs(self.args.args_d['initvar'][0])
    else:
      G[0,0,0] = panel.args.init_var
    
    #Idea for IV: calculate Z*u throughout. Mazimize total sum of LL. 
    u = panel.Y-fu.dot(X,self.args.args_d['beta'])
    u_RE = (u+self.re_obj_i.RE(u, panel)+self.re_obj_t.RE(u, panel))*incl

    matrices=self.arma_calc(panel, u_RE, self.h_add, G)
    if matrices is None:
      return None		
    AMA_1,AMA_1AR,GAR_1,GAR_1MA, e_RE, var, h=matrices

    #NOTE: self.h_val itself is also set in ctypes.cpp/ctypes.c. If you change self.h_val below, you need to 
    #change it in the c-scripts too. self.h_val must be calcualted below as well for later calulcations. 
    if panel.options.EGARCH.value==0:
      e_REsq =(e_RE**2+(e_RE==0)*1e-18) 
      nd =1
      self.h_val, self.h_e_val, self.h_2e_val = (e_RE**2+self.h_add)*incl, nd*2*e_RE*incl, nd*2*incl
      self.h_z_val, self.h_2z_val,  self.h_ez_val = None,None,None	#possibility of additional parameter in sq res function		
    else:
      minesq = 1e-20
      e_REsq =np.maximum(e_RE**2,minesq)
      nd = e_RE**2>minesq		

      self.h_val, self.h_e_val, self.h_2e_val = np.log(e_REsq+self.h_add)*incl, 2*incl*e_RE/(e_REsq+self.h_add), incl*2/(e_REsq+self.h_add) - incl*2*e_RE**2/(e_REsq+self.h_add)**2
      self.h_z_val, self.h_2z_val,  self.h_ez_val = None,None,None	#possibility of additional parameter in sq res function		

    self.variance_RE(panel,e_REsq)
    if False:#debug
      from .. import debug
      if np.any(h!=self.h_val):
        print('the h calculated in the c function and the self.h_val calcualted here do not match')
      debug.test_c_armas(u_RE, var, e_RE, panel, self, G)

    LL_full,v,v_inv,self.dvar_pos=function.LL(panel,var,e_REsq, e_RE, self.minvar, self.maxvar)
    self.tobit(panel,LL_full)
    LL=np.sum(LL_full*incl)
    self.LL_all=np.sum(LL_full)
    self.add_variables(panel,matrices, u, u_RE, var, v, G,e_RE,e_REsq,v_inv,LL_full)
    if abs(LL)>1e+100: 
      return None				
    return LL

  def set_var_bounds(self, panel):
    if panel.options.EGARCH.value==0:
      self.minvar = 0.01*panel.args.init_var
      self.maxvar = 1000*panel.args.init_var
      self.h_add = panel.args.init_var
    else:
      self.minvar = -100
      self.maxvar = 100
      self.h_add = 0.1
      
  def add_variables(self,panel,matrices,u, u_RE,var,v,G,e_RE,e_REsq,v_inv,LL_full):
    self.v_inv05=v_inv**0.5
    self.e_norm=e_RE*self.v_inv05	
    self.e_RE_norm_centered=(self.e_norm-panel.mean(self.e_norm))*panel.included[3]
    self.u, self.u_RE      = u,  u_RE
    self.var,  self.v,    self.LL_full = var,       v,    LL_full
    self.G=G
    self.e_RE=e_RE
    self.e_REsq=e_REsq
    self.v_inv=v_inv

  def tobit(self,panel,LL):
    if sum(panel.tobit_active)==0:
      return
    g=[1,-1]
    self.F=[None,None]	
    for i in [0,1]:
      if panel.tobit_active[i]:
        I=panel.tobit_I[i]
        self.F[i]= stat_dist.norm(g[i]*self.e_norm[I])
        LL[I]=np.log(self.F[i])


  def variance_RE(self,panel,e_REsq):
    """Calculates random/fixed effects for variance."""
    #not in use, expermental. Should be applied to normalize before ARIMA/GARCH
    self.vRE,self.varRE,self.dvarRE=panel.zeros[3],panel.zeros[3],panel.zeros[3]
    self.ddvarRE,self.dvarRE_mu,self.ddvarRE_mu_vRE=panel.zeros[3],None,None
    self.varRE_input, self.ddvarRE_input, self.dvarRE_input = None, None, None
    return
    if panel.options.fixed_random_variance_eff.value==0:
      return panel.zeros[3]
    if panel.N==0:
      return None

    meane2=panel.mean(e_REsq)
    self.varRE_input=(e_REsq-meane2)*panel.included[3]

    mine2=0
    mu=panel.options.variance_RE_norm.value
    self.vRE_i=self.re_obj_i_v.RE(self.varRE_input, panel)
    self.vRE_t=self.re_obj_t_v.RE(self.varRE_input, panel)
    self.meane2=meane2
    vRE=meane2*panel.included[3]-self.vRE_i-self.vRE_t
    self.vRE=vRE
    small=vRE<=mine2
    big=small==False
    vREbig=vRE[big]
    vREsmall=vRE[small]

    varREbig=np.log(vREbig+mu)
    varREsmall=(np.log(mine2+mu)+((vREsmall-mine2)/(mine2+mu)))
    varRE,dvarRE,ddvarRE=np.zeros(vRE.shape),np.zeros(vRE.shape),np.zeros(vRE.shape)

    varRE[big]=varREbig
    varRE[small]=varREsmall
    self.varRE=varRE*panel.included[3]

    dvarRE[big]=1/(vREbig+mu)
    dvarRE[small]=1/(mine2+mu)
    self.dvarRE=dvarRE*panel.included[3]

    ddvarRE[big]=-1/(vREbig+mu)**2
    self.ddvarRE=ddvarRE*panel.included[3]

    return self.varRE



  def standardize(self,panel,reverse_difference=False):
    """Adds X and Y and error terms after ARIMA-E-GARCH transformation and random effects to self. 
    If reverse_difference and the ARIMA difference term d>0, the standardized variables are converted to
    the original undifferenced order. This may be usefull if the predicted values should be used in another 
    differenced regression."""
    if hasattr(self,'Y_st'):
      return		
    m=panel.lost_obs
    N,T,k=panel.X.shape
    if model_parser.DEFAULT_INTERCEPT_NAME in panel.args.caption_d['beta']:
      m=self.args.args_d['beta'][0,0]
    else:
      m=panel.mean(panel.Y)	
    #e_norm=self.standardize_variable(panel,self.u,reverse_difference)
    self.Y_long = panel.input.Y
    self.X_long = panel.input.X
    self.Y_st,   self.Y_st_long   = self.standardize_variable(panel,panel.Y,reverse_difference)
    self.X_st,   self.X_st_long   = self.standardize_variable(panel,panel.X,reverse_difference)
    self.XIV_st, self.XIV_st_long = self.standardize_variable(panel,panel.XIV,reverse_difference)
    self.Y_pred_st=fu.dot(self.X_st,self.args.args_d['beta'])
    self.Y_pred=fu.dot(panel.X,self.args.args_d['beta'])	
    self.e_RE_norm_centered_long=self.stretch_variable(panel,self.e_RE_norm_centered)
    self.Y_pred_st_long=self.stretch_variable(panel,self.Y_pred_st)
    self.Y_pred_long=np.dot(panel.input.X,self.args.args_d['beta'])
    self.u_long=np.array(panel.input.Y-self.Y_pred_long)
    
    a=0


  def standardize_variable(self,panel,X,norm=False,reverse_difference=False):
    X=panel.arma_dot.dot(self.AMA_1AR,X,self)
    X=(X+self.re_obj_i.RE(X, panel,False)+self.re_obj_t.RE(X, panel,False))
    if (not panel.Ld_inv is None) and reverse_difference:
      X=fu.dot(panel.Ld_inv,X)*panel.a[3]		
    if norm:
      X=X*self.v_inv05
    X_long=self.stretch_variable(panel,X)
    return X,X_long		

  def stretch_variable(self,panel,X):
    N,T,k=X.shape
    m=panel.map
    NT=panel.total_obs
    X_long=np.zeros((NT,k))
    X_long[m]=X
    return X_long



  def copy_args_d(self):
    return copy_array_dict(self.args.args_d)


  def h(self,panel,e,z):
    return h(e, z, panel)

  def arma_calc(self,panel, u, h_add, G):
    matrices = arma.set_garch_arch(panel,self.args.args_d, u, h_add, G)
    if matrices is None:
      return None		
    self.AMA_1,self.AMA_1AR,self.GAR_1,self.GAR_1MA, self.e, self.var, self.h = matrices
    self.AMA_dict={'AMA_1':None,'AMA_1AR':None,'GAR_1':None,'GAR_1MA':None}		
    return matrices
  
  def predict(self, W, W_next = None):
    d = self.args.args_d
    self.u_pred = pred_u(self.u, self.e, d['rho'], d['lambda'])
    #u_pred = pred_u(self.u[:,:-1], self.e[:,:-1], d['rho'], d['lambda'], self.e[:,-1])#test
    self.var_pred = pred_var(self.h, self.var, d['psi'], d['gamma'], d['omega'], W_next, self.minvar, self.maxvar)
    #var_pred = pred_var(self.h[:,:-1], self.var[:,:-1], d['psi'], d['gamma'], d['omega'], W, self.minvar, self.maxvar)#test
    if not hasattr(self,'Y_pred'):
      self.standardize()
    
    return {'predicted residual':self.u_pred, 'predicted variance':self.var_pred, 
            'in-sample predicted Y': self.Y_pred_long, 'in-sample predicted variance': self.v}

    
def pred_u(u, e, rho, lmbda, e_now = 0):
  if len(lmbda)==0 and len(rho)==0:
    return 0
  u_pred = e_now
  if len(rho):
    u_pred += sum([
      rho[i]*u[:,-i-1] for i in range(len(rho))
      ])
  if len(lmbda):
    u_pred += sum([
      lmbda[i]*e[:,-i-1] for i in range(len(lmbda))
    ])  
  if len(u_pred)==1:
    u_pred = u_pred[0,0]
  return u_pred
  
def pred_var(h, var, psi, gamma, omega, W, minvar, maxvar):
  W = test_variance_signal(W, h, omega)
  if W is None:
    G =omega[0,0]
  else:
    G = np.dot(W,omega)
  a, b = 0, 0 
  if len(psi):
    a = sum([
      psi[i]*h[:,-i-1] for i in range(len(psi))
      ])
  if len(gamma):
    b = sum([
      gamma[i]*(var[:,-i-1]) for i in range(len(gamma))
    ])  
    
  var_pred = G + a +b
  var_pred = np.maximum(np.minimum(var_pred, maxvar), minvar)
  if len(var_pred)==1:
    var_pred = var_pred[0,0]  
  return var_pred



def test_variance_signal(W, h, omega):
  if W is None:
    return None
  N,T,k= h.shape
  if N==1:
    W = W.flatten()
    if len(W)!=len(omega):
        raise RuntimeError("The variance signals needs to be a numpy array of numbers with "
                           "the size equal to the HF-argument variables +1, and first variable must be 1")
    return W.reshape((1,len(omega)))
  else:
    try:
      NW,kW = W.shape
      if NW!=N or kW!=k:
        raise RuntimeError("Rows and columns in variance signals must correspond with"
                           "the number of groups and the size equal to the number of "
                           "HF-argument variables +1, respectively")
    except:
      raise RuntimeError("If there are more than one group, the variance signal must be a matrix with"
                         "Rows and columns in variance signals must correspond with"
                         "the number of groups and the size equal to the number of "
                           "HF-argument variables +1, respectively"                       )      
  return W
  





def add_to_matrices(X_1,X_1b,a,ab,r):
  for i in range(0,len(a)):	
    if i>0:
      d=(r[i:],r[:-i])
      X_1[d]=a[i]
    else:
      d=(r,r)
    X_1b[d]=ab[i]	
  return X_1,X_1b

def lag_matr(L,args):
  k=len(args)
  if k==0:
    return L
  L=1*L
  r=np.arange(len(L))
  for i in range(k):
    d=(r[i+1:],r[:-i-1])
    if i==0:
      d=(r,r)
    L[d]=args[i]

  return L



def copy_array_dict(d):
  r=dict()
  for i in d:
    r[i]=np.array(d[i])
  return r


#!/usr/bin/env python
# -*- coding: utf-8 -*-

from . import linesearch
from . import direction
from . import constraints
from .. import likelihood_simple as logl
import numpy as np
import time

#This module finds the array of arguments that minimizes some function. The derivative 
#of the function also needs to be supplied. 
#This is an adaption of the Broyden-Fletcher-Goldfarb-Shanno variant of Davidon-Fletcher-Powell algorithm by 
#Press, William H,Saul A Teukolsky,William T Vetterling and Brian P Flannery. 1992. Numerical Recipes in C'. 
#Cambridge: Cambridge University Press.



EPS=3.0e-16 
TOLX=(4*EPS) 
GTOL = 0.001

def dfpmax(x, f, g, hessin, H, panel, slave_id):
  """Given a starting point x[1..n] that is a vector of length n, the Broyden-Fletcher-Goldfarb-
  Shanno variant of Davidon-Fletcher-Powell minimization is performed on a function func, using
  its gradient as calculated by a routine dfunc. The convergence requirement on zeroing the
  gradient is input as gtol. Returned quantities are x[1..n] (the location of the minimum),
  iter (the number of iterations that were performed), and fret (the minimum value of the
  function). The routine lnsrch is called to perform approximate line minimizations.
  fargs are fixed arguments that ar not subject to optimization. ("Nummerical Recipes for C") """


  ll = logl.LL(x, panel)
  
  its, msg = 0, ''
  MAXITER = 10000

  for its in range(MAXITER):  	#Main loop over the iterations.
    constr = get_constr(its, H, ll, x, panel)
    dx, dx_norm, H_ = direction.get(g, x, H, constr, f, hessin, simple=False)
    ls = linesearch.LineSearch(x, panel, ll)
    ls.lnsrch(x, f, g, H, dx, constr)	

    dx_realized = ls.x - x
    incr = ls.f - f

    ll = ls.ll

    x, f, hessin, H, g, conv= calc(dx_realized,  hessin, H, incr, its, ls, panel, constr)

    err = (np.max(np.abs(dx_realized)) < TOLX) and its>5

    if conv==1:
      msg = "Convergence on zero gradient; local or global minimum identified"
    elif conv==2:
      msg = "Convergence on zero expected gain; local or global minimum identified given multicolinearity constraints"		
    elif conv==3:
      msg = "Reached the maximum number of iterations"		  
    elif err:
      msg = "Warning: Convergence on delta x; the gradient is incorrect or the tolerance is set too low"
    elif its>MAXITER:
      msg = "No convergence within %s iterations" %(MAXITER,)


    if (conv>0) or err or its+1==MAXITER:	
      return msg, conv, ls.ll.args.args_v

def get_constr(its,H, ll, x, panel):
  constr = constraints.Constraints(panel, x)
  constr.add_static_constraints(panel, its, ll,  np.nonzero(ll.var<ll.minvar)[1])	
  constr.add_dynamic_constraints(panel, H, ll)	
  return constr

def calc(dx_realized,hessin, H, incr, its, ls, panel, constr):
  f, x, g_old, rev, alam,ll = ls.f, ls.x, ls.g, ls.rev, ls.alam, ls.ll
  #Thhese setting may not hold for all circumstances, and should be tested properly:

  g, G = calc_gradient(ll, panel)
  
  a = np.ones(len(g))
  if not constr is None:
    a[list(constr.fixed.keys())] =0		
    a[ls.applied_constraints] = 0    
  g_norm =np.max(np.abs(g*a*x)/(abs(f)+1e-12) )
  gtol = GTOL
  if sum(a)==1:
    gtol = 1e-10

  if abs(g_norm) < gtol:
    return x, f, hessin, H, g, 1
  
  print(f"its:{its}, f:{f}, gnorm: {abs(g_norm)}")

  hessin=hessin_num(hessin, g-g_old, dx_realized)
  H = inv(hessin, H)

  return x, f, hessin, H, g, 0

  
def calc_gradient(ll, panel):
  dLL_lnv, DLL_e=logl.func_gradent(ll,panel)
  g, G = logl.gradient(ll, panel,DLL_e,dLL_lnv,return_G=True)
  return g, G



def hessin_num(hessin, dg, xi):				#Compute difference of gradients,
  n=len(dg)
  #and difference times current matrix:
  hdg=(np.dot(hessin,dg.reshape(n,1))).flatten()
  fac=fae=sumdg=sumxi=0.0 							#Calculate dot products for the denominators. 
  fac = np.sum(dg*xi) 
  fae = np.sum(dg*hdg)
  sumdg = np.sum(dg*dg) 
  sumxi = np.sum(xi*xi) 
  if (fac < (EPS*sumdg*sumxi)**0.5):  					#Skip update if fac not sufficiently positive.
    fac=1.0/fac 
    fad=1.0/fae 
                            #The vector that makes BFGS different from DFP:
    dg=fac*xi-fad*hdg   
    #The BFGS updating formula:
    hessin+=fac*xi.reshape(n,1)*xi.reshape(1,n)
    hessin-=fad*hdg.reshape(n,1)*hdg.reshape(1,n)
    hessin+=fae*dg.reshape(n,1)*dg.reshape(1,n)		

  return hessin




def inv(hessian, H):
  try:
    return np.linalg.inv(hessian)
  except:
    return H



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import numpy as np
from datetime import datetime
from datetime import date
from . import model_parser
import pandas as pd
import json
import csv
forbidden_names=['tobit_low','tobit_high',model_parser.DEFAULT_INTERCEPT_NAME,model_parser.CONST_NAME]

def load(filepath_or_buffer,sep=None, header="infer", 
         names=None, index_col=None, usecols=None, squeeze=False, prefix=None, 
                                         mangle_dupe_cols=True, dtype=None, engine=None, converters=None, 
                                         true_values=None, false_values=None, skipinitialspace=False, skiprows=None, 
                                         skipfooter=0, nrows=None, na_values=None, keep_default_na=True, na_filter=True, 
                                         verbose=False, skip_blank_lines=True, parse_dates=False, infer_datetime_format=False, 
                                         keep_date_col=False, date_parser=None, dayfirst=False, cache_dates=True, 
                                         iterator=False, chunksize=None, compression="infer", thousands=None, decimal=".", 
                                         lineterminator=None, quotechar='"', quoting=csv.QUOTE_MINIMAL, doublequote=True, 
                                         escapechar=None, comment=None, encoding=None, dialect=None, error_bad_lines=True, 
                                         warn_bad_lines=True, delim_whitespace=False, 
                                         memory_map=False, float_precision=None):
  print ("opening file ...")
  sep=get_sep(filepath_or_buffer, sep)
  data=pd.read_csv(filepath_or_buffer, sep, sep,header, 
                         names, index_col, usecols, squeeze, prefix, 
                                         mangle_dupe_cols, dtype, engine, converters, 
                                         true_values, false_values, skipinitialspace, skiprows, 
                                         skipfooter, nrows, na_values, keep_default_na, na_filter, 
                                         verbose, skip_blank_lines, parse_dates, infer_datetime_format, 
                                         keep_date_col, date_parser, dayfirst, cache_dates, 
                                         iterator, chunksize, compression, thousands, decimal, 
                                         lineterminator, quotechar, quoting, doublequote, 
                                         escapechar, comment, encoding, dialect, error_bad_lines, 
                                         warn_bad_lines, delim_whitespace,
                                         memory_map=memory_map, float_precision=float_precision)
  print ("... done")
  load_data_printout(data)
  return data

def load_json(path_or_buf=None, orient=None, typ="frame", dtype=None, convert_axes=None, convert_dates=True, 
              keep_default_dates=True, numpy=False, precise_float=False, date_unit=None, encoding=None, 
                          lines=False, chunksize=None, compression="infer"):
  return pd.read_json(path_or_buf, orient, typ, dtype, convert_axes, convert_dates, 
                            keep_default_dates, numpy, precise_float, date_unit, 
                                 encoding, lines, chunksize, compression)


def load_data_printout(data):
  print ("The following variables were loaded:"+', '.join(data.keys()))


def append(d,key,i):
  if key in d:
    d[key].append(i)
  else:
    d[key]=[i]

def load_SQL(sql_string,conn,index_col=None, coerce_float=True, params=None, parse_dates=None, columns=None, chunksize=None):
  return pd.read_sql(sql_string, conn, index_col, coerce_float, params, parse_dates, columns, chunksize)




def get_sep(fname,sep):
  f=open(fname,'r')
  r=[]
  sample_size=20
  for i in range(sample_size):
    r.append(f.readline())	
  f.close()
  d={}
  for i in [sep,';',',','\t',' ']:#checks whether the separator is consistent
    len0=len(r[0].split(i))
    err=False
    for j in r[1:]:
      rlen=len(j.split(i))
      if rlen!=len0:
        err=True
        break
    if not err and rlen>1:
      d[i]=rlen
  maxlen=max([d[i] for i in d])
  if ',' in d:
    if d[',']==maxlen:
      if len(d)>0:
        d.pop('.')
        maxlen=max([d[i] for i in d])
  for i in d:
    if d[i]==maxlen:
      return i

SQL_type_dict={0: float,
               1: int,
 2: int,
 3: float,
 4: float,
 5: float,
 6: float,
 8: int,
 9: int,
 16: int,
 246: int
 }


#!/usr/bin/env python
# -*- coding: utf-8 -*-

import numpy as np
from .. import functions as fu



def dd_func_lags_mult(panel,ll,g,AMAL,vname1,vname2,transpose=False, u_gradient=False):
  de2_zeta_xi_RE,de2_zeta_xi 	= dd_func_lags_mult_arima(panel,ll,g,AMAL,vname1,vname2,transpose, u_gradient)
  dd_re_variance 							= dd_func_re_variance(panel,ll,g,vname1,vname2,de2_zeta_xi_RE,u_gradient)
  d2LL_d2var_zeta_xi,d2LL_d2e_zeta_xi_RE 	= dd_func_garch(panel,ll,g,vname1,vname2,de2_zeta_xi_RE,de2_zeta_xi,dd_re_variance,u_gradient)

  return d2LL_d2var_zeta_xi,d2LL_d2e_zeta_xi_RE

def dd_func_lags_mult_arima(panel,ll,g,AMAL,vname1,vname2,transpose, u_gradient):
  #de_xi is "N x T x m", de_zeta is "N x T x k" and L is "T x T"
  de_xi=g.__dict__['de_'+vname1+'_RE']
  de_zeta=g.__dict__['de_'+vname2+'_RE']
  de2_zeta_xi_RE=None
  if de_xi is None or de_zeta is None:
    return None,None
  (N,T,m)=de_xi.shape
  (N,T,k)=de_zeta.shape	
  #ARIMA:
  if not AMAL is None:
    if u_gradient:
      de2_zeta_xi=-panel.arma_dot.dot(AMAL,g.X_RE,ll)#"T x N x s x m #for error beta-rho covariance, the u gradient must be used	
    else:
      de2_zeta_xi=panel.arma_dot.dot(AMAL,de_zeta,ll)#"T x N x s x m
    if transpose:#only happens if lags==k
      de2_zeta_xi=de2_zeta_xi+np.swapaxes(de2_zeta_xi,2,3)#adds the transpose
    de2_zeta_xi=de2_zeta_xi*panel.included[4]
  else:
    de2_zeta_xi=None
  de2_zeta_xi_RE=de2_zeta_xi

  return de2_zeta_xi_RE,de2_zeta_xi


def dd_func_re_variance(panel,ll,g,vname1,vname2,de2_zeta_xi_RE,u_gradient):
  if panel.N<=1 or panel.options.fixed_random_group_eff.value==0:
    return None
  #voaltility RE:

  de_xi_RE=g.__dict__['de_'+vname1+'_RE']
  de_zeta_RE=g.__dict__['de_'+vname2+'_RE']
  if de_xi_RE is None or de_zeta_RE is None:
    return None
  (N,T,m)=de_xi_RE.shape
  (N,T,k)=de_zeta_RE.shape		
  de_xi_RE_r=de_xi_RE.reshape((N,T,m,1))
  de_zeta_RE_r=de_zeta_RE.reshape((N,T,1,k))	
  incl=panel.included[4]

  dvRE_xi=g.__dict__['dvRE_'+vname1].reshape((N,T,m,1))
  dvRE_zeta=g.__dict__['dvRE_'+vname2].reshape((N,T,1,k))			

  if not de2_zeta_xi_RE is None:
    d_xi_input=g.__dict__['d_'+vname1+'_input']
    d_zeta_input=g.__dict__['d_'+vname2+'_input']		
    try:
      dd_e_RE_sq=(2*de_xi_RE_r*de_zeta_RE_r+2*ll.e_RE.reshape(N,T,1,1)*de2_zeta_xi_RE)*incl
    except:
      dd_e_RE_sq=(2*de_xi_RE_r*de_zeta_RE_r+2*ll.e_RE.reshape(N,T,1,1)*de2_zeta_xi_RE)*incl
    ddmeane2= panel.mean(dd_e_RE_sq,(0,1))*incl
    dd_input=(dd_e_RE_sq-ddmeane2)*incl

    ddvRE_d_xi_zeta=ddmeane2-add((ll.re_obj_i_v.ddRE(dd_input,d_xi_input,d_zeta_input,ll.varRE_input,vname1,vname2,panel),
                                              ll.re_obj_t_v.ddRE(dd_input,d_xi_input,d_zeta_input,ll.varRE_input,vname1,vname2,panel)),True)
  else:
    ddvRE_d_xi_zeta=None

  dvarRE,		ddvarRE		=	ll.dvarRE.reshape((N,T,1,1)),		ll.ddvarRE.reshape((N,T,1,1))
  dd_re_variance	=	add((prod((dvarRE,ddvRE_d_xi_zeta))	,	ddvarRE*dvRE_xi*dvRE_zeta))
  return dd_re_variance


def dd_func_garch(panel,ll,g,vname1,vname2,de2_zeta_xi_RE,de2_zeta_xi,dd_re_variance,u_gradient):
  #GARCH: 
  incl=panel.included[4]
  if (g.__dict__['de_'+vname1+'_RE'] is None) or (g.__dict__['de_'+vname2+'_RE'] is None):
    return None, None	
  (N,T,m)=g.__dict__['de_'+vname1+'_RE'].shape
  (N,T,k)=g.__dict__['de_'+vname2+'_RE'].shape
  DLL_e=g.DLL_e.reshape(N,T,1,1)

  d2LL_d2e_zeta_xi_RE=None
  if not de2_zeta_xi_RE is None:	
    d2LL_d2e_zeta_xi_RE = de2_zeta_xi_RE * DLL_e	
    d2LL_d2e_zeta_xi_RE = np.sum(np.sum(d2LL_d2e_zeta_xi_RE*incl,0),0)

  RE_suffix=''

  de_xi_RE=g.__dict__['de_'+vname1+'_RE'+RE_suffix]
  de_zeta_RE=g.__dict__['de_'+vname2+'_RE'+RE_suffix]	


  dLL_var=g.dLL_var.reshape(N,T,1,1)
  if de2_zeta_xi_RE is None:
    de2_zeta_xi_RE=0
  d2var_zeta_xi=None
  d2LL_d2var_zeta_xi=None
  d_omega_e=None
  d2var_zeta_xi_h=None
  if panel.pqdkm[4]>0:
    if u_gradient:
      de_zeta_RE=g.__dict__['de_'+vname2+'_RE'+RE_suffix]
    h_e_de2_zeta_xi =  ll.h_e_val.reshape(N,T,1,1)  * de2_zeta_xi_RE
    h_2e_dezeta_dexi = ll.h_2e_val.reshape(N,T,1,1) * de_xi_RE.reshape((N,T,m,1)) * de_zeta_RE.reshape((N,T,1,k))

    d2var_zeta_xi_h = (h_e_de2_zeta_xi + h_2e_dezeta_dexi)

    d2var_zeta_xi_h = panel.arma_dot.dot(ll.GAR_1MA, d2var_zeta_xi_h,ll)

  d2var_zeta_xi = add((d2var_zeta_xi_h,  dd_re_variance,d_omega_e), True)
  if not d2var_zeta_xi is None:
    d2LL_d2var_zeta_xi = np.sum(prod((d2var_zeta_xi,dLL_var*incl),True),(0,1))

  return d2LL_d2var_zeta_xi,d2LL_d2e_zeta_xi_RE

def dd_func_lags(panel,ll,L,d,dLL,transpose=False):
  #d is "N x T x m" and L is "k x T x T"
  if panel.pqdkm[4]==0:
    return None
  if d is None:
    return None		
  (N,T,m)=d.shape
  if L is None:
    x=0
  elif len(L)==0:
    return None
  elif type(L)==tuple:#elif len(L.shape)==3:
    x=panel.arma_dot.dot(L,d,ll)#"T x N x k x m"
    if x is None:
      return None
  elif len(L.shape)==2:
    x=fu.dot(L,d).reshape(N,T,1,m)
  dLL=dLL.reshape((N,T,1,1))
  return np.sum(np.sum(dLL*x,1),0)#and sum it	


def add(iterable,ignore=False):
  """Sums iterable. If ignore=True all elements except those that are None are added. If ignore=False, None is returned if any element is None. """
  x=None
  for i in range(len(iterable)):
    if not iterable[i] is None:
      if x is None:
        x=iterable[i]
      else:
        x=x+iterable[i]
    else:
      if not ignore:
        return None
  return x

def prod(iterable,ignore=False):
  """Takes the product sum of iterable. If ignore=True all elements except those that are None are multiplied. 
  If ignore=False, None is returned if any element is None. """
  x=None
  for i in iterable:
    if not i is None:
      if x is None:
        x=i
      else:
        x=x*i
    else:
      if not ignore:
        return None
  return x

def sumNT(nparray):
  if nparray is None:
    return None
  s=nparray.shape
  if len(s)<3:
    raise RuntimeError("Not enough dimensions")

  return np.sum(nparray.reshape(list(s)+[1]),(0,1))


def concat_matrix(block_matrix):
  m=[]
  for i in range(len(block_matrix)):
    r=block_matrix[i]
    C=[]
    for j in range(len(r)):
      if not r[j] is None:
        C.append(r[j])
    if len(C):
      m.append(np.concatenate(C,1))
  m=np.concatenate(m,0)
  return m

def concat_marray(matrix_array):
  arr=[]
  for i in matrix_array:
    if not i is None:
      arr.append(i)
  arr=np.concatenate(arr,2)
  return arr

def dd_func(d2LL_de2,d2LL_dln_de,d2LL_dln2,de_dh,de_dg,dln_dh,dln_dg,dLL_de2_dh_dg,dLL_dln2_dh_dg):
  a=[]
  a.append(dd_func_mult(de_dh,d2LL_de2,de_dg))

  a.append(dd_func_mult(de_dh,d2LL_dln_de,dln_dg))
  a.append(dd_func_mult(dln_dh,d2LL_dln_de,de_dg))

  a.append(dd_func_mult(dln_dh,d2LL_dln2,dln_dg))

  a.append(dLL_de2_dh_dg)
  a.append(dLL_dln2_dh_dg)
  return add(a,True)

def dd_func_mult(d0,mult,d1):
  #d0 is N x T x k and d1 is N x T x m
  if d0 is None or d1 is None or mult is None:
    return None
  (N,T,k)=d0.shape
  (N,T,m)=d1.shape
  if np.any(np.isnan(d0)) or np.any(np.isnan(d1)):
    x=np.empty((k,m))
    x[:]=np.nan
    return x
  d0=d0*mult
  d0=np.reshape(d0,(N,T,k,1))
  d1=np.reshape(d1,(N,T,1,m))
  try:
    x=np.sum(np.sum(d0*d1,0),0)#->k x m 
  except RuntimeWarning as e:
    if e.args[0]=='overflow encountered in multiply':
      d0=np.minimum(np.maximum(d0,-1e+100),1e+100)
      d1=np.minimum(np.maximum(d1,-1e+100),1e+100)
      x=np.sum(np.sum(d0*d1,0),0)#->k x m 
    else:
      raise RuntimeWarning(e)
  return x





#!/usr/bin/env python
# -*- coding: utf-8 -*-

import numpy as np
from .. import functions as fu


def add(iterable,ignore=False):
  """Sums iterable. If ignore=True all elements except those that are None are added. If ignore=False, None is returned if any element is None. """
  x=None
  for i in range(len(iterable)):
    if not iterable[i] is None:
      if x is None:
        x=iterable[i]
      else:
        x=x+iterable[i]
    else:
      if not ignore:
        return None
  return x

def prod(iterable,ignore=False):
  """Takes the product sum of iterable. If ignore=True all elements except those that are None are multiplied. 
  If ignore=False, None is returned if any element is None. """
  x=None
  for i in iterable:
    if not i is None:
      if x is None:
        x=i
      else:
        x=x*i
    else:
      if not ignore:
        return None
  return x

def sumNT(nparray):
  if nparray is None:
    return None
  s=nparray.shape
  if len(s)<3:
    raise RuntimeError("Not enough dimensions")

  return np.sum(nparray.reshape(list(s)+[1]),(0,1))


def concat_matrix(block_matrix):
  m=[]
  for i in range(len(block_matrix)):
    r=block_matrix[i]
    C=[]
    for j in range(len(r)):
      if not r[j] is None:
        C.append(r[j])
    if len(C):
      m.append(np.concatenate(C,1))
  m=np.concatenate(m,0)
  return m

def concat_marray(matrix_array):
  arr=[]
  for i in matrix_array:
    if not i is None:
      arr.append(i)
  arr=np.concatenate(arr,2)
  return arr

def dd_func(d2LL_de2,d2LL_dln_de,d2LL_dln2,de_dh,de_dg,dln_dh,dln_dg,dLL_de2_dh_dg,dLL_dln2_dh_dg):
  a=[]
  a.append(dd_func_mult(de_dh,d2LL_de2,de_dg))

  a.append(dd_func_mult(de_dh,d2LL_dln_de,dln_dg))
  a.append(dd_func_mult(dln_dh,d2LL_dln_de,de_dg))

  a.append(dd_func_mult(dln_dh,d2LL_dln2,dln_dg))

  a.append(dLL_de2_dh_dg)
  a.append(dLL_dln2_dh_dg)
  return add(a,True)

def dd_func_mult(d0,mult,d1):
  #d0 is N x T x k and d1 is N x T x m
  if d0 is None or d1 is None or mult is None:
    return None
  (N,T,k)=d0.shape
  (N,T,m)=d1.shape
  if np.any(np.isnan(d0)) or np.any(np.isnan(d1)):
    x=np.empty((k,m))
    x[:]=np.nan
    return x
  d0=d0*mult
  d0=np.reshape(d0,(N,T,k,1))
  d1=np.reshape(d1,(N,T,1,m))
  try:
    x=np.sum(np.sum(d0*d1,0),0)#->k x m 
  except RuntimeWarning as e:
    if e.args[0]=='overflow encountered in multiply':
      d0=np.minimum(np.maximum(d0,-1e+100),1e+100)
      d1=np.minimum(np.maximum(d1,-1e+100),1e+100)
      x=np.sum(np.sum(d0*d1,0),0)#->k x m 
    else:
      raise RuntimeWarning(e)
  return x





#!/usr/bin/env python
# -*- coding: utf-8 -*-

import os
from pathlib import Path
import numpy as np
import numpy.ctypeslib as npct
import ctypes as ct
p = os.path.join(Path(__file__).parent.absolute(),'cfunctions')
if os.name=='nt':
  cfit = npct.load_library('cfit.dll',p)
else:
  cfit = npct.load_library('cfit.so',p)
  
CDPT = ct.POINTER(ct.c_double) 
CIPT = ct.POINTER(ct.c_uint) 
  
n = 5
x = np.random.rand(n, n)
y = np.zeros((n,n))

def main():
  
  cfit.fit(x.ctypes.data_as(CIPT), y.ctypes.data_as(CIPT),n)		
  print(y)
  
main()


#!/usr/bin/env python
# -*- coding: utf-8 -*-

from ..output import stat_functions as stat
import numpy as np


def get(g, x, H, constr, f, hessin, simple=True):
  n = len(x)
  if simple or (H is None):
    dx = -(np.dot(hessin,g.reshape(n,1))).flatten()
  else:
    dx, H, applied_constraints = solve(constr, H, g, x, f)	
  dx_norm = normalize(dx, x)

  return dx, dx_norm, H

def new(g, x, H, constr, f,dx, alam):
  dx, slope, rev = slope_check(g, dx)
  if constr is None:
    return dx*alam, slope, rev, []
  elif len(constr.intervals)==0:
    return dx*alam, slope, rev, []
  elif constr.within(x + dx):
    return dx*alam, slope, rev, []

  dxalam, H, applied_constraints = solve(constr, H, g*alam, x, f)	
  if np.sum(g*dxalam)<0.0:
    dxalam, H, applied_constraints = solve(constr, H, -g*alam, x, f)	
    slope = np.sum(g*dx)
    rev = True
  return dxalam, slope, rev, applied_constraints

def slope_check(g, dx):
  rev = False
  slope=np.sum(g*dx)					#Scale if attempted step is too big.
  if slope <= 0.0:
    dx=-dx
    slope=np.sum(g*dx)
    rev = True  
  return dx, slope, rev

def solve(constr,H, g, x, f):
  """Solves a second degree taylor expansion for the dc for df/dc=0 if f is quadratic, given gradient
  g, hessian H, inequalty constraints c and equalitiy constraints c_eq and returns the solution and 
  and index constrained indicating the constrained variables"""

  if H is None:
    raise RuntimeError('Cant solve with no coefficient matrix')
  try:
    list(constr.keys())[0]
  except:
    dx = -np.linalg.solve(H, g)
    return dx, H, []
  H = make_hess_posdef(H)
  H_orig = np.array(H)
  m=len(H)

  idx=np.ones(m,dtype=bool)
  delmap=np.arange(m)
  if len(list(constr.fixed.keys()))>0:#removing fixed constraints from the matrix
    idx[list(constr.fixed.keys())]=False
    H=H[idx][:,idx]
    g=g[idx]
    delmap-=np.cumsum(idx==False)
    delmap[idx==False]=m#if for some odd reason, the deleted variables are referenced later, an out-of-bounds error will be thrown
  n=len(H)
  keys=list(constr.intervals.keys())
  k=len(keys)
  H=np.concatenate((H,np.zeros((n,k))),1)
  H=np.concatenate((H,np.zeros((k,n+k))),0)
  g=np.append(g,np.zeros(k))

  for i in range(k):
    H[n+i,n+i]=1
  
  dx = -np.linalg.solve(H, g)
  dx_full=np.zeros(m)
  OK=False
  
  applied_constraints=[]
  for j in range(k):
    dx_full[idx]=dx[:n]
    OK=constr.within(x+dx_full,False)
    if OK:break 
    key=keys[j]
    dx, H = kuhn_tucker(constr,key,j,n, H, g, x, f,dx,delmap, OK)
    applied_constraints.append(key)
  dx_full=np.zeros(m)
  dx_full[idx]=dx[:n]
  H_full = np.zeros((m,m))
  idx = idx.reshape((1,m))
  nz = np.nonzero(idx*idx.T)
  H_full[nz] = H[np.nonzero(np.ones((n,n)))]
  H_full[H_full==0] = H_orig[H_full==0]
  return dx_full, H_full, applied_constraints


def kuhn_tucker(constr,key,j,n,H,g,x, f, dx,delmap, OK,recalc=True):
  q=None
  c=constr.intervals[key]
  i=delmap[key]
  if not c.value is None:
    q=-(c.value-x[key])
  elif x[key]+dx[i]<c.min:
    q=-(c.min-x[key])
  elif x[key]+dx[i]>c.max:
    q=-(c.max-x[key])
  if q!=None:
    if OK:
      a=0
    H[i,n+j]=1
    H[n+j,i]=1
    H[n+j,n+j]=0
    g[n+j]=q
    if recalc:
      dx = -np.linalg.solve(H, g)
  return dx, H

def normalize(dx, x):
  x = np.abs(x)
  dx_norm=(x!=0)*dx/(x+(x<1e-100))
  dx_norm=(x<1e-2)*dx+(x>=1e-2)*dx_norm	
  return dx_norm	

def make_hess_posdef(H):
  c, var_prop, ev, p = stat.var_decomposition(XXNorm=H)
  ev = ((ev<0) - 1.0 * (ev>=0)) * ev
  H = np.dot(p, np.dot(np.diag(ev), p.T))  
  return H




#!/usr/bin/env python
# -*- coding: utf-8 -*-

#calculates the arma matrices for GARCH and ARIMA

from pathlib import Path
import os
import numpy.ctypeslib as npct
import ctypes as ct
p = os.path.join(Path(__file__).parent.absolute(),'cfunctions')
if os.name=='nt':
  cfunct = npct.load_library('ctypes.dll',p)
else:
  cfunct = npct.load_library('ctypes.so',p)
import numpy as np


CDPT = ct.POINTER(ct.c_double) 
CIPT = ct.POINTER(ct.c_uint) 


def set_garch_arch(panel,args,u, h_add, G):
  """Solves X*a=b for a where X is a banded matrix with 1 or zero, and args along
  the diagonal band"""
  N, T, _ = u.shape
  rho=np.insert(-args['rho'],0,1)
  psi=args['psi']
  psi=np.insert(args['psi'],0,0) 



  lmbda = args['lambda']
  gmma = -args['gamma']
  
  parameters = np.array(( N , T , 
                  len(lmbda), len(rho), len(gmma), len(psi), 
                  panel.options.EGARCH.value, panel.lost_obs, 
                  h_add))
  if True:
    AMA_1,AMA_1AR,GAR_1,GAR_1MA, e, var, h = inv_c(parameters, lmbda, rho, gmma, psi, N, T, u, G, panel.T_arr)
  else:
    AMA_1,AMA_1AR,GAR_1,GAR_1MA, e, var, h = inv_python(parameters, lmbda, rho, gmma, psi, N, T, u, G, panel.T_arr)

  r=[]
  #Creating nympy arrays with name properties. 
  for i in ['AMA_1','AMA_1AR','GAR_1','GAR_1MA']:
    r.append((locals()[i],i))
  for i in ['e', 'var', 'h']:
    r.append(locals()[i])

  return r

def inv_c(parameters,lmbda, rho,gmma, psi, N, T, u, G, T_arr):
  AMA_1,AMA_1AR,GAR_1,GAR_1MA, e, var, h=(
          np.append([1],np.zeros(T-1)),
                np.zeros(T),
                np.append([1],np.zeros(T-1)),
                np.zeros(T),
                np.zeros((N,T,1)),
                np.zeros((N,T,1)),
                np.zeros((N,T,1))
        )

  T_arr = np.array(T_arr.flatten(),dtype = float)
  cfunct.armas(parameters.ctypes.data_as(CIPT), 
                     lmbda.ctypes.data_as(CDPT), rho.ctypes.data_as(CDPT),
                                                  gmma.ctypes.data_as(CDPT), psi.ctypes.data_as(CDPT),
                                                  AMA_1.ctypes.data_as(CDPT), AMA_1AR.ctypes.data_as(CDPT),
                                                  GAR_1.ctypes.data_as(CDPT), GAR_1MA.ctypes.data_as(CDPT),
                                                  u.ctypes.data_as(CDPT), 
                                                  e.ctypes.data_as(CDPT), 
                                                  var.ctypes.data_as(CDPT),
                                                  h.ctypes.data_as(CDPT),
                                                  G.ctypes.data_as(CDPT), 
                                                  T_arr.ctypes.data_as(CIPT)
                                                  )		
  return AMA_1,AMA_1AR,GAR_1,GAR_1MA, e, var, h
  
def inv_python(parameters,lmbda, rho,gmma, psi, N, T, u, G, T_arr):
  AMA_1,AMA_1AR,GAR_1,GAR_1MA, e, var, h=(
      np.append([1],np.zeros(T-1)),
            np.zeros(T),
            np.append([1],np.zeros(T-1)),
            np.zeros(T),
            np.zeros(N*T),
            np.zeros(N*T),
            np.zeros(N*T)
    )

  u = u.reshape(N*T)
  G = G.reshape(N*T)
  armas(parameters,lmbda, rho,gmma, psi,
         AMA_1, AMA_1AR,GAR_1, GAR_1MA,u, e, 
         var,h,G, T_arr.flatten())		
  
  e, var, h, u, G=(e.reshape((N,T,1)),
            var.reshape((N,T,1)),
            h.reshape((N,T,1)), 
            u.reshape((N,T,1)), 
            G.reshape((N,T,1)))  
  return AMA_1,AMA_1AR,GAR_1,GAR_1MA, e, var, h

def inverse(n, x_args, nx, b_args, nb, a, ab):
  a[0] = 1.0
  ab[0] = b_args[0]

  for i in range(1, n):
    sum_ax = 0
    sum_ab = 0
    for j in range(min(i, nx)):
      sum_ax += x_args[j] * a[i-j-1]
    a[i] = -sum_ax
    for j in range(min(i+1, nb)):
      sum_ab += b_args[j] * a[i-j]
    ab[i] = sum_ab

  return a, ab

def armas(parameters, lambda_, rho, gamma, psi, AMA_1, AMA_1AR, GAR_1, GAR_1MA, u, e, var, h, W, T_arr):
  N = int(parameters[0])
  T = int(parameters[1])
  nlm = int(parameters[2])
  nrh = int(parameters[3])
  ngm = int(parameters[4])
  npsi = int(parameters[5])
  egarch = int(parameters[6])
  lost_obs = int(parameters[7])
  h_add = parameters[8]

  inverse(T, lambda_, nlm, rho, nrh, AMA_1, AMA_1AR)
  inverse(T, gamma, ngm, psi, npsi, GAR_1, GAR_1MA)

  for k in range(N):  # individual dimension
    for i in range(T_arr[k]):  # time dimension TODO: the upper limit here should be the maximum date for each group. 
      # ARMA:
      sum_e = 0
      for j in range(i+1):  # time dimension, backtracking
        sum_e += AMA_1AR[j] * u[(i-j) + k*T]
      e[i + k*T] = sum_e
      # GARCH:
      if i >= lost_obs:
        h[i + k*T] = sum_e**2 + h_add
        if egarch:
          h[i + k*T] = np.log(h[i + k*T] + 1e-18 * (h[i + k*T] == 0))
      sum_var = 0
      for j in range(i+1):  # time dimension, backtracking
        sum_var += GAR_1[j] * W[(i-j) + k*T] + GAR_1MA[j]*h[(i-j) + k*T]
      var[i + k*T] = sum_var

  return 0

  
def set_garch_arch_scipy(panel,args):
  #after implementing ctypes, the scipy version might be dropped entirely
  p,q,d,k,m=panel.pqdkm
  nW,n=panel.nW,panel.max_T

  AAR=-lag_matr(-panel.I,args['rho'])
  AMA_1AR,AMA_1=solve_mult(args['lambda'], AAR, panel.I)
  if AMA_1AR is None:
    return
  GMA=lag_matr(panel.I*0,args['psi'])
  GAR_1MA,GAR_1=solve_mult(-args['gamma'], GMA, panel.I)
  if GAR_1MA is None:
    return
  r=[]
  for i in ['AMA_1','AMA_1AR','GAR_1','GAR_1MA']:
    r.append((locals()[i],i))
  return r

def solve_mult(args,b,I):
  """Solves X*a=b for a where X is a banded matrix with 1  and args along
  the diagonal band"""
  import scipy
  n=len(b)
  q=len(args)
  X=np.zeros((q+1,n))
  X[0,:]=1
  X2=np.zeros((n,n))
  w=np.zeros(n)
  r=np.arange(n)	
  for i in range(q):
    X[i+1,:n-i-1]=args[i]
  try:
    X_1=scipy.linalg.solve_banded((q,0), X, I)
    if np.any(np.isnan(X_1)):
      return None,None			
    X_1b=fu.dot(X_1, b)
  except:
    return None,None

  return X_1b,X_1


#!/usr/bin/env python
# -*- coding: utf-8 -*-
import numpy as np
MIN_DEN = 0

def LL(panel,var,e_REsq, e_RE, minvar, maxvar):
  incl=panel.included[3]

  LL_const=-0.5*np.log(2*np.pi)
  if panel.options.EGARCH.value==0:
    a,k=panel.options.GARCH_assist.value, panel.options.kurtosis_adj.value

    dvar_pos=(var<maxvar)*(var>minvar) 
    var = incl*np.maximum(np.minimum(var,maxvar),minvar)
    v=var	
    v_inv = incl/(var + MIN_DEN + (incl==0))	

    LL_full = LL_const-0.5*(incl*np.log(var+MIN_DEN + (incl==0))+(1-k)*e_REsq*v_inv
                                        + a* (np.abs(e_REsq-var)*v_inv)
                                                                + (k/3)* e_REsq**2*v_inv**2
                                                                )
  else:
    dvar_pos=(var < maxvar) * (var > minvar)
    var = np.maximum(np.minimum(var, maxvar), minvar)
    v = np.exp(var)*incl
    v_inv = np.exp(-var)*incl		
    LL_full = LL_const-0.5*(var+(e_REsq)*v_inv)
  return LL_full,v,v_inv,dvar_pos


def gradient(ll,panel):
  incl=panel.included[3]
  a,k=panel.options.GARCH_assist.value, panel.options.kurtosis_adj.value
  var,e_REsq,e_RE,v_inv=ll.var, ll.e_REsq, ll.e_RE,ll.v_inv 

  if panel.options.EGARCH.value==0:
    DLL_e   =-0.5*(	(1-k)*2*e_RE*v_inv	)
    dLL_var =-0.5*(	v_inv-(1-k)*(e_REsq)*v_inv**2	)

    DLL_e +=-0.5*(		
                  a* 2*np.sign(e_REsq-var)*e_RE*v_inv
                          + (k/3)* 4*e_REsq*e_RE*v_inv**2
                )
    dLL_var +=-0.5*(	
                  - a* (np.sign(e_REsq-var)*v_inv)
                                - a* (np.abs(e_REsq-var)*v_inv**2)
                                                - (k/3)* 2*e_REsq**2*v_inv**3
                )
  else:	
    DLL_e=-(ll.e_RE*ll.v_inv)
    dLL_var=-0.5*(incl-(ll.e_REsq*ll.v_inv)*incl)	
  dLL_var*=ll.dvar_pos*incl	
  DLL_e*=incl
  return dLL_var, DLL_e

def hessian(ll,panel):
  incl=panel.included[3]
  var,e_REsq,e_RE,v_inv=ll.var, ll.e_REsq, ll.e_RE,ll.v_inv 
  a,k=panel.options.GARCH_assist.value, panel.options.kurtosis_adj.value

  if panel.options.EGARCH.value==0:	
    d2LL_de2 	=-0.5*(	(1-k)*2*v_inv	)
    d2LL_dln_de =-0.5*(	-(1-k)*2*e_RE*v_inv**2)
    d2LL_dln2 	=-0.5*(-1*v_inv**2+(1-k)*2*(e_REsq)*v_inv**3	)

    d2LL_de2 	+=-0.5*(		
                  a* 2*np.sign(e_REsq-var)*v_inv
                                  + (k/3)* 12*e_REsq*v_inv**2
                )
    d2LL_dln_de +=-0.5*(		
                  - a* 2*np.sign(e_REsq-var)*e_RE*v_inv**2
                                - (k/3)* 8*e_REsq*e_RE*v_inv**3
                )
    d2LL_dln2 	+=-0.5*(	
                  + a* (np.sign(e_REsq-var)*v_inv**2)
                                        + a* 2*(np.abs(e_REsq-var)*v_inv**3)
                                                        + a* (np.sign(e_REsq-var)*v_inv**2)
                                                        + (k/3)* 6*e_REsq**2*v_inv**4
                )

  else:
    d2LL_de2=-ll.v_inv
    d2LL_dln_de=ll.e_RE*ll.v_inv
    d2LL_dln2=-0.5*ll.e_REsq*ll.v_inv
  d2LL_dln_de*=ll.dvar_pos*incl
  d2LL_dln2*=ll.dvar_pos*incl
  d2LL_de2*=incl

  return d2LL_de2, d2LL_dln_de, d2LL_dln2




#!/usr/bin/env python
# -*- coding: utf-8 -*-

from . import dfpmax
from . import computation
from ..likelihood_simple import dfpmax as dfpmax_smpl


from .. import likelihood as logl
import numpy as np
from ..parallel import callback




def maximize(args, inbox, outbox, panel, gtol, tolx, nummerical, diag_hess, slave_id):
  args = np.array(args)
  callbk = callback.CallBack(inbox, outbox)
  comput = computation.Computation(args, panel, gtol, tolx, None, nummerical, diag_hess)
  

  callbk.callback(quit=False, conv=False, perc=0)
  initval = InitialValues(panel, comput)
  
  x, ll, f, g, hessin, H = initval.calc_init_dir(args)
  res, ll = dfpmax.dfpmax(x, f, g, hessin, H, comput, callbk, panel, slave_id, ll)
  #msg, conv, args_c = dfpmax_smpl.dfpmax(x, f, g, hessin, H, panel, slave_id)

  return res, ll


class InitialValues:
  def __init__(self, panel, comput):
    self.comput = comput
    self.panel = panel
    

  def init_ll(self,args=None, ll=None):
    if args is None:
      args = ll.args.args_v
    
    try:
      args=args.args_v
    except:
      pass#args must be a vector
    for i in self.comput.constr.fixed:
      args[i]=self.comput.constr.fixed[i].value
    if ll is None:
      ll=logl.LL(args, self.panel, constraints=self.comput.constr,print_err=True)
    if ll.LL is None:
      print("WARNING: Initial arguments failed, attempting default OLS-arguments ...")
      self.panel.args.set_init_args(self.panel,default=True)
      ll=logl.LL(self.panel.args.args_OLS,self.panel,constraints=self.constr,print_err=True)
      if ll.LL is None:
        raise RuntimeError("OLS-arguments failed too, you should check the data")
      else:
        print("default OLS-arguments worked")
    return ll
  
  def calc_init_dir(self, p0, full = False, diag_hess = False):
    """Calculates the initial computation""" 
    ll = self.init_ll(p0)
    g, G = self.comput.calc_gradient(ll)
    if self.panel.options.use_analytical.value==2:
      H = -np.identity(len(g))
      hessin = H
      return p0, ll, ll.LL , g, hessin, H
    H = self.comput.calc_hessian(ll)
    if diag_hess:
      d = np.diag(H)
      d = d - (np.abs(d)<1e-100)
      H = np.diag(d)
      hessin = np.diag(1/d)
    else:
      hessin = np.linalg.inv(H)
    return p0, ll, ll.LL , g, hessin, H


#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""A setuptools based setup module.
See:
https://packaging.python.org/en/latest/distributing.html
https://github.com/pypa/sampleproject
"""

version='1.2.36'

from setuptools import setup, find_packages
from codecs import open
from os import path

# Get the long description from the README file
here = path.abspath(path.dirname(__file__))
f = open(path.join(here, 'README.txt'), encoding='utf-8')
long_description = f.read()

setup(
    name='paneltime',
    version=version,
    description='An efficient integrated panel and GARCH estimator',
    long_description=long_description,
    url='https://github.com/espensirnes/paneltime',
    author='Espen Sirnes',
    author_email='espen.sirnes@uit.no',
    license='GPL-3.0',

    classifiers=[
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        'Development Status :: 4 - Beta',
        'Intended Audience :: Science/Research',
        'Topic :: Scientific/Engineering',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Programming Language :: Python :: 3.8',
        ],

  keywords='econometrics',

  packages=find_packages(exclude=['contrib', 'docs', 'tests']),

  install_requires=['numpy >= 1.11','pymysql', 'pandas',  'mpmath'],
	extras_require={'linux':'gcc'},	

  package_data={
      '': ['*.ico','likelihood/cfunctions/*'],
      },
  include_package_data=True,

  entry_points={
      'console_scripts': [
          'paneltime=paneltime:main',
          ],
      },
)




#!/usr/bin/env python
# -*- coding: utf-8 -*-


from .. import functions as fu
from . import calculus_functions as cf
from . import function
import numpy as np
import time


def arima_grad(k,x,ll,sign,pre, panel):
  if k==0:
    return None
  (N,T,m)=x.shape
  x=panel.arma_dot.dotroll(pre,k,sign,x,ll)
  x.resize(N,T,k)
  extr_value=1e+100
  if np.max(np.abs(x))>extr_value:
    x[np.abs(x)>extr_value]=np.sign(x[np.abs(x)>extr_value])*extr_value
  return x*panel.a[3]

def garch_arima_grad(ll,d,varname, panel):
  panel=panel
  dvar_sigma=None
  if panel.pqdkm[4]>0 and not d is None: 			#eqs. 33-34
    ((N,T,k))=d.shape
    x=cf.prod((ll.h_e_val,d))	
    dvar_sigma=panel.arma_dot.dot(ll.GAR_1MA,x,ll)
  return dvar_sigma


def gradient(ll, panel,DLL_e=None,dLL_var=None,return_G=False):

  incl= panel.included[3]

  u, e, h_e_val,var,h_val,v=ll.u, ll.e,ll.h_e_val,ll.var,ll.h_val,ll.v
  p,q,d,k,m=panel.pqdkm

  if DLL_e is None:
    dLL_var, DLL_e=function.gradient(ll,panel)
  #ARIMA:
  de_rho=arima_grad(p,u*incl,ll,-1,ll.AMA_1, panel)
  de_lambda=arima_grad(q,e*incl,ll,-1,ll.AMA_1, panel)
  de_beta=-panel.arma_dot.dot(ll.AMA_1AR,panel.XIV*incl,ll)*panel.a[3]

  (de_rho,de_lambda,de_beta)=(de_rho,de_lambda,de_beta)		

  dvar_sigma_rho		=	garch_arima_grad(ll,	de_rho,		'rho', panel)
  dvar_sigma_lambda	=	garch_arima_grad(ll,	de_lambda,	'lambda', panel)
  dvar_sigma_beta		=	garch_arima_grad(ll,	de_beta,	'beta', panel)


  #GARCH:

  dvar_omega=panel.arma_dot.dot(ll.GAR_1,panel.W_a,ll)
  dvar_initvar = None
  if 'initvar' in ll.args.args_d:
    dvar_initvar = ll.GAR_1[0].reshape((1,panel.max_T,1))
    if not panel.options.EGARCH.value and ll.args.args_d['initvar'][0]<0:	
      dvar_initvar = -dvar_initvar
    
  (dvar_gamma, dvar_psi, dvar_mu, dvar_z_G, dvar_z)=(None,None,None,None,None)
  
  if panel.N>1:
    dvar_mu=cf.prod((ll.dvarRE_mu,incl))
  else:
    dvar_mu=None	

  if m>0:
    dvar_gamma=arima_grad(k,var,ll,1,ll.GAR_1, panel)
    dvar_psi=arima_grad(m,h_val,ll,1,ll.GAR_1, panel)
    if not ll.h_z_val is None:
      dvar_z_G=fu.dot(ll.GAR_1MA,ll.h_z_val)
      (N,T,k)=dvar_z_G.shape

    dvar_z=dvar_z_G


  (dvar_gamma, dvar_psi,dvar_mu,dvar_z_G,dvar_z)=(dvar_gamma, dvar_psi, dvar_mu, dvar_z_G, dvar_z)

  #LL


  #final derivatives:
  dLL_beta=cf.add((cf.prod((dvar_sigma_beta,dLL_var)),cf.prod((de_beta,DLL_e))),True)
  dLL_rho=cf.add((cf.prod((dvar_sigma_rho,dLL_var)),cf.prod((de_rho,DLL_e))),True)
  dLL_lambda=cf.add((cf.prod((dvar_sigma_lambda,dLL_var)),cf.prod((de_lambda,DLL_e))),True)
  dLL_gamma=cf.prod((dvar_gamma,dLL_var))
  dLL_psi=cf.prod((dvar_psi,dLL_var))
  dLL_omega=cf.prod((dvar_omega,dLL_var))
  dLL_initvar=cf.prod((dvar_initvar,dLL_var))
  dLL_mu=cf.prod((dvar_mu,dLL_var))
  dLL_z=cf.prod((dvar_z,dLL_var))


  G=cf.concat_marray((dLL_beta,dLL_rho,dLL_lambda,dLL_gamma,dLL_psi,dLL_omega, dLL_initvar,dLL_mu,dLL_z))
  g=np.sum(G,(0,1))
  #For debugging:
  #from .. import debug
  #print(debug.grad_debug(ll,panel,0.00001))
  #print(g)
  #if np.sum((g-gn)**2)>10000000:
  #	a=0
  #print(gn)
  #a=debug.grad_debug_detail(ll, panel, 0.00000001, 'LL', 'beta',0)
  #dLLeREn,deREn=debug.LL_calc_custom(ll, panel, 0.0000001)


  if return_G:
    return  g,G
  else:	
    return g
